@@ -24,10 +24,11 @@ except NameError:
 
 
 class SelectorError(Exception):
-    """Invalid or unsupported selector.
+    """Common parent for :class:`SelectorSyntaxError` and
+    :class:`ExpressionError`.
 
-    Common parent class for the exceptions that are actually raised.
-    You can use it to catch any error in a selector.
+    You can just use ``except SelectorError:`` when calling
+    :meth:`~GenericTranslator.css_to_xpath` and handle both exceptions types.
 
     """
 
@@ -37,63 +38,116 @@ class SelectorSyntaxError(SelectorError, SyntaxError):
 
 #### Parsed objects
 
+class Selector(object):
+    """
+    Represents a selector with an optional pseudo element.
+    """
+    def __init__(self, tree, pseudo_element=None):
+        self._tree = tree
+        #: If the selector has a pseudo-element: a string like ``'after'``.
+        #: Otherwise, ``None``.
+        #: Any identifier preceded by ``::`` is accepted as a pseudo-element.
+        #: It is the user’s responsibility to reject selectors with
+        #: unknown or unsupported pseudo-elements.
+        self.pseudo_element = pseudo_element
+
+    def __repr__(self):
+        if self.pseudo_element:
+            pseudo_element = '::%s' % self.pseudo_element
+        else:
+            pseudo_element = ''
+        return '%s[%r%s]' % (
+            self.__class__.__name__, self._tree, pseudo_element)
+
+    def specificity(self):
+        """Return the specificity_ of this selector as a tuple of 3 integers.
+
+        .. _specificity: http://www.w3.org/TR/selectors/#specificity
+
+        """
+        a, b, c = self._tree.specificity()
+        if self.pseudo_element:
+            c += 1
+        return a, b, c
+
+
 class Class(object):
     """
     Represents selector.class_name
     """
-
     def __init__(self, selector, class_name):
         self.selector = selector
         self.class_name = class_name
 
     def __repr__(self):
         return '%s[%r.%s]' % (
-            self.__class__.__name__,
-            self.selector,
-            self.class_name)
+            self.__class__.__name__, self.selector, self.class_name)
+
+    def specificity(self):
+        a, b, c = self.selector.specificity()
+        b += 1
+        return a, b, c
 
 
 class Function(object):
     """
     Represents selector:name(expr)
     """
-
-    def __init__(self, selector, type, name, arguments):
+    def __init__(self, selector, name, arguments):
         self.selector = selector
-        self.type = type  # TODO: is this needed?
         self.name = name
         self.arguments = arguments
 
     def __repr__(self):
-        return '%s[%r%s%s(%r)]' % (
-            self.__class__.__name__,
-            self.selector,
-            self.type, self.name, self.arguments)
+        return '%s[%r:%s(%r)]' % (
+            self.__class__.__name__, self.selector, self.name, self.arguments)
+
+    def specificity(self):
+        a, b, c = self.selector.specificity()
+        b += 1
+        return a, b, c
 
 
 class Pseudo(object):
     """
     Represents selector:ident
     """
-
-    def __init__(self, selector, type, ident):
+    def __init__(self, selector, ident):
         self.selector = selector
-        assert type in (':', '::')
-        self.type = type
         self.ident = ident
 
     def __repr__(self):
-        return '%s[%r%s%s]' % (
-            self.__class__.__name__,
-            self.selector,
-            self.type, self.ident)
+        return '%s[%r:%s]' % (
+            self.__class__.__name__, self.selector, self.ident)
+
+    def specificity(self):
+        a, b, c = self.selector.specificity()
+        b += 1
+        return a, b, c
+
+
+class Negation(object):
+    """
+    Represents selector:not(subselector)
+    """
+    def __init__(self, selector, subselector):
+        self.selector = selector
+        self.subselector = subselector
+
+    def __repr__(self):
+        return '%s[%r:not(%r)]' % (
+            self.__class__.__name__, self.selector, self.subselector)
+
+    def specificity(self):
+        a1, b1, c1 = self.selector.specificity()
+        a2, b2, c2 = self.sub_selector.specificity()
+        return a1 + a2, b1 + b2, c1 + c2
 
 
 class Attrib(object):
     """
     Represents selector[namespace|attrib operator value]
     """
-
     def __init__(self, selector, namespace, attrib, operator, value):
         self.selector = selector
         self.namespace = namespace
@@ -102,74 +156,66 @@ class Attrib(object):
         self.value = value
 
     def __repr__(self):
+        if self.namespace == '*':
+            attrib = self.attrib
+        else:
+            attrib = '%s|%s' % (self.namespace, self.attrib)
         if self.operator == 'exists':
             return '%s[%r[%s]]' % (
-                self.__class__.__name__,
-                self.selector,
-                self._format_attrib())
+                self.__class__.__name__, self.selector, attrib)
         else:
             return '%s[%r[%s %s %r]]' % (
-                self.__class__.__name__,
-                self.selector,
-                self._format_attrib(),
-                self.operator,
-                self.value)
+                self.__class__.__name__, self.selector, attrib,
+                self.operator, self.value)
 
-    def _format_attrib(self):
-        if self.namespace == '*':
-            return self.attrib
-        else:
-            return '%s|%s' % (self.namespace, self.attrib)
+    def specificity(self):
+        a, b, c = self.selector.specificity()
+        b += 1
+        return a, b, c
 
 
 class Element(object):
     """
     Represents namespace|element
     """
-
     def __init__(self, namespace, element):
         self.namespace = namespace
         self.element = element
 
     def __repr__(self):
-        return '%s[%s]' % (
-            self.__class__.__name__,
-            self._format_element())
-
-    def _format_element(self):
         if self.namespace == '*':
-            return self.element
+            element = self.element
         else:
-            return '%s|%s' % (self.namespace, self.element)
+            element = '%s|%s' % (self.namespace, self.element)
+        return '%s[%s]' % (
+            self.__class__.__name__, element)
+
+    def specificity(self):
+        if self.element == '*':
+            return 0, 0, 0
+        else:
+            return 0, 0, 1
 
 
 class Hash(object):
     """
     Represents selector#id
     """
-
     def __init__(self, selector, id):
         self.selector = selector
         self.id = id
 
     def __repr__(self):
         return '%s[%r#%s]' % (
-            self.__class__.__name__,
-            self.selector, self.id)
+            self.__class__.__name__, self.selector, self.id)
 
-
-class Or(object):
-
-    def __init__(self, items):
-        self.items = items
-    def __repr__(self):
-        return '%s(%r)' % (
-            self.__class__.__name__,
-            self.items)
+    def specificity(self):
+        a, b, c = self.selector.specificity()
+        a += 1
+        return a, b, c
 
 
 class CombinedSelector(object):
-
     def __init__(self, selector, combinator, subselector):
         assert selector is not None
         self.selector = selector
@@ -182,10 +228,12 @@ class CombinedSelector(object):
         else:
             comb = self.combinator
         return '%s[%r %s %r]' % (
-            self.__class__.__name__,
-            self.selector,
-            comb,
-            self.subselector)
+            self.__class__.__name__, self.selector, comb, self.subselector)
+
+    def specificity(self):
+        a1, b1, c1 = self.selector.specificity()
+        a2, b2, c2 = self.subselector.specificity()
+        return a1 + a2, b1 + b2, c1 + c2
 
 
 #### Parser
@@ -195,22 +243,38 @@ _id_re = re.compile(r'^\s*(\w*)#(\w+)\s*$')
 _class_re = re.compile(r'^\s*(\w*)\.(\w+)\s*$')
 
 
-def parse(string):
-    # Fast path for simple cases
-    match = _el_re.match(string)
-    if match:
-        return Element('*', match.group(1))
-    match = _id_re.match(string)
-    if match is not None:
-        return Hash(Element('*', match.group(1) or '*'), match.group(2))
-    match = _class_re.match(string)
-    if match is not None:
-        return Class(Element('*', match.group(1) or '*'), match.group(2))
+def parse(css):
+    """Parse a CSS *group of selectors*.
 
-    stream = TokenStream(tokenize(string))
-    stream.source = string
+    If you don't care about pseudo-elements or selector specificity,
+    you can skip this and use :meth:`~GenericTranslator.css_to_xpath`.
+
+    :param css:
+        A *group of selectors* as an Unicode string.
+    :raises:
+        :class:`SelectorSyntaxError` on invalid selectors.
+    :returns:
+        A list of parsed :class:`Selector` objects, one for each
+        selector in the comma-separated group.
+
+    """
+    # Fast path for simple cases
+    match = _el_re.match(css)
+    if match:
+        return [Selector(Element('*', match.group(1)))]
+    match = _id_re.match(css)
+    if match is not None:
+        return [Selector(Hash(Element(
+            '*', match.group(1) or '*'), match.group(2)))]
+    match = _class_re.match(css)
+    if match is not None:
+        return [Selector(Class(Element(
+            '*', match.group(1) or '*'), match.group(2)))]
+
+    stream = TokenStream(tokenize(css))
+    stream.source = css
     try:
-        return parse_selector_group(stream)
+        return list(parse_selector_group(stream))
     except SelectorSyntaxError:
         import sys
         e = sys.exc_info()[1]
@@ -224,29 +288,26 @@ def parse(string):
 
 
 def parse_selector_group(stream):
-    result = []
     stream.skip_whitespace()
     while 1:
-        result.append(parse_selector(stream))
+        yield Selector(*parse_selector(stream))
         if stream.peek() == ',':
             stream.next()
             stream.skip_whitespace()
         else:
             break
-    if len(result) == 1:
-        return result[0]
-    else:
-        return Or(result)
-
 
 def parse_selector(stream):
-    result = parse_simple_selector(stream)
+    result, pseudo_element = parse_simple_selector(stream)
     while 1:
         stream.skip_whitespace()
         peek = stream.peek()
         if peek == ',' or peek is None:
-            return result
-        elif peek in ('+', '>', '~'):
+            break
+        if pseudo_element:
+            raise SelectorSyntaxError(
+                'A pseudo-element must be at the end of a selector')
+        if peek in ('+', '>', '~'):
             # A combinator
             combinator = stream.next()
             stream.skip_whitespace()
@@ -254,9 +315,9 @@ def parse_selector(stream):
             # By exclusion, the last parse_simple_selector() ended
             # at peek == ' '
             combinator = ' '
-        next_selector = parse_simple_selector(stream)
+        next_selector, pseudo_element = parse_simple_selector(stream)
         result = CombinedSelector(result, combinator, next_selector)
-    return result
+    return result, pseudo_element
 
 
 def parse_simple_selector(stream, inside_negation=False):
@@ -275,8 +336,15 @@ def parse_simple_selector(stream, inside_negation=False):
     else:
         element = namespace = '*'
     result = Element(namespace, element)
+    pseudo_element = None
     while 1:
         peek = stream.peek()
+        if peek in (None, ' ', ',', '+', '>', '~') or (
+                inside_negation and peek == ')'):
+            break
+        if pseudo_element:
+            raise SelectorSyntaxError(
+                'A pseudo-element must be at the end of a selector')
         if peek == '#':
             stream.next()
             result = Hash(result, stream.next_symbol())
@@ -293,17 +361,29 @@ def parse_simple_selector(stream, inside_negation=False):
                 raise SelectorSyntaxError(
                     "] expected, got '%s'" % next)
             continue
-        elif peek == ':' or peek == '::':
-            type = stream.next()
+        elif peek == '::':
+            stream.next()
+            pseudo_element = stream.next_symbol()
+            continue
+        elif peek == ':':
+            stream.next()
             ident = stream.next_symbol()
+            if ident in ('first-line', 'first-letter', 'before', 'after'):
+                # Special case: CSS 2.1 pseudo-elements can have a single ':'
+                # Any new pseudo-element must have two.
+                pseudo_element = ident
+                continue
             if stream.peek() == '(':
                 stream.next()
                 stream.skip_whitespace()
                 if ident == 'not':
                     if inside_negation:
-                        raise SyntaxError('Got nested :not()')
-                    argument = parse_simple_selector(
+                        raise SelectorSyntaxError('Got nested :not()')
+                    argument, argument_pseudo_element = parse_simple_selector(
                         stream, inside_negation=True)
+                    if argument_pseudo_element:
+                        raise SelectorSyntaxError(
+                            'Pseudo-elements are not allowed inside :not()')
                 else:
                     peek = stream.peek()
                     if isinstance(peek, (Symbol, String)):
@@ -316,20 +396,20 @@ def parse_simple_selector(stream, inside_negation=False):
                 if not next == ')':
                     raise SelectorSyntaxError(
                         "Expected ')', got '%s'" % next)
-                result = Function(result, type, ident, argument)
+                if ident == 'not':
+                    result = Negation(result, argument)
+                else:
+                    result = Function(result, ident, argument)
             else:
-                result = Pseudo(result, type, ident)
+                result = Pseudo(result, ident)
             continue
-        elif peek in (None, ' ', ',', '+', '>', '~') or (
-                inside_negation and peek == ')'):
-            break
         else:
             raise SelectorSyntaxError(
                 "Expected selector, got '%s'" % peek)
     if consumed == len(stream.used):
         raise SelectorSyntaxError(
             "Expected selector, got '%s'" % stream.peek())
-    return result
+    return result, pseudo_element
 
 
 def parse_attrib(selector, stream):
@@ -539,7 +619,6 @@ def tokenize_symbol(s, pos):
 
 
 class TokenStream(object):
-
     def __init__(self, tokens, source=None):
         self.used = []
         self.tokens = iter(tokens)
