@@ -21,8 +21,9 @@ import operator
 import unittest
 
 from lxml import html
-from cssselect.parser import tokenize, parse, parse_series, SelectorSyntaxError
-from cssselect.xpath import GenericTranslator, HTMLTranslator, ExpressionError
+from cssselect import (parse, GenericTranslator, HTMLTranslator,
+                       SelectorSyntaxError, ExpressionError)
+from cssselect.parser import tokenize, parse_series
 
 
 class TestCssselect(unittest.TestCase):
@@ -42,78 +43,157 @@ class TestCssselect(unittest.TestCase):
             "Token(']', 15)"]
 
     def test_parser(self):
-        def parse_many(first, *others):
-            result = repr(parse(first))
-            for other in others:
-                assert repr(parse(other)) == result
-            return result.replace("(u'", "('")
+        def repr_parse(css):
+            selectors = parse(css)
+            for selector in selectors:
+                assert selector.pseudo_element is None
+            return [repr(selector._tree).replace("(u'", "('")
+                    for selector in selectors]
 
-        assert parse_many('*') == 'Element[*]'
-        assert parse_many('*|*') == 'Element[*]'
-        assert parse_many('*|foo') == 'Element[foo]'
-        assert parse_many('foo|*') == 'Element[foo|*]'
-        assert parse_many('foo|bar') == 'Element[foo|bar]'
+        def parse_many(first, *others):
+            result = repr_parse(first)
+            for other in others:
+                assert repr_parse(other) == result
+            return result
+
+        assert parse_many('*') == ['Element[*]']
+        assert parse_many('*|*') == ['Element[*]']
+        assert parse_many('*|foo') == ['Element[foo]']
+        assert parse_many('foo|*') == ['Element[foo|*]']
+        assert parse_many('foo|bar') == ['Element[foo|bar]']
         # This will never match, but it is valid:
-        assert parse_many('#foo#bar') == 'Hash[Hash[Element[*]#foo]#bar]'
+        assert parse_many('#foo#bar') == ['Hash[Hash[Element[*]#foo]#bar]']
         assert parse_many(
             'div>.foo',
             'div> .foo',
             'div >.foo',
             'div > .foo',
             'div \n>  \t \t .foo', 'div\r>\n\n\n.foo', 'div\f>\f.foo'
-        ) == 'CombinedSelector[Element[div] > Class[Element[*].foo]]'
+        ) == ['CombinedSelector[Element[div] > Class[Element[*].foo]]']
         assert parse_many('td.foo,.bar',
             'td.foo, .bar',
             'td.foo\t\r\n\f ,\t\r\n\f .bar'
-        ) == 'Or([Class[Element[td].foo], Class[Element[*].bar]])'
-        assert parse_many('div, td.foo, div.bar span') == (
-            'Or([Element[div], Class[Element[td].foo], '
-                'CombinedSelector[Class[Element[div].bar] '
-                '<followed> Element[span]]])')
-        assert parse_many('div > p') == (
-            'CombinedSelector[Element[div] > Element[p]]')
-        assert parse_many('td:first') == (
-            'Pseudo[Element[td]:first]')
-        assert parse_many('td::first') == (
-            'Pseudo[Element[td]::first]')
-        assert parse_many('td :first') == (
+        ) == [
+            'Class[Element[td].foo]',
+            'Class[Element[*].bar]'
+        ]
+        assert parse_many('div, td.foo, div.bar span') == [
+            'Element[div]',
+            'Class[Element[td].foo]',
+            'CombinedSelector[Class[Element[div].bar] '
+                '<followed> Element[span]]']
+        assert parse_many('div > p') == [
+            'CombinedSelector[Element[div] > Element[p]]']
+        assert parse_many('td:first') == [
+            'Pseudo[Element[td]:first]']
+        assert parse_many('td:first') == [
+            'Pseudo[Element[td]:first]']
+        assert parse_many('td :first') == [
             'CombinedSelector[Element[td] '
-                '<followed> Pseudo[Element[*]:first]]')
-        assert parse_many('td ::first') == (
+                '<followed> Pseudo[Element[*]:first]]']
+        assert parse_many('td :first') == [
             'CombinedSelector[Element[td] '
-                '<followed> Pseudo[Element[*]::first]]')
-        assert parse_many('a[name]', 'a[ name\t]') == (
-            'Attrib[Element[a][name]]')
-        assert parse_many('a [name]') == (
-            'CombinedSelector[Element[a] <followed> Attrib[Element[*][name]]]')
-        assert parse_many('a[rel="include"]') == (
-            "Attrib[Element[a][rel = String('include', 6)]]")
-        assert parse_many('a[rel = include]') == (
-            "Attrib[Element[a][rel = Symbol('include', 8)]]")
-        assert parse_many("a[hreflang |= 'en']") == (
-            "Attrib[Element[a][hreflang |= String('en', 14)]]")
-        assert parse_many('div:nth-child(10)') == (
-            "Function[Element[div]:nth-child(Symbol('10', 14))]")
-        assert parse_many(':nth-child(2n+2)') == (
-            "Function[Element[*]:nth-child(Symbol('2n+2', 11))]")
-        assert parse_many('div:nth-of-type(10)') == (
-            "Function[Element[div]:nth-of-type(Symbol('10', 16))]")
-        assert parse_many('div div:nth-of-type(10) .aclass') == (
+                '<followed> Pseudo[Element[*]:first]]']
+        assert parse_many('a[name]', 'a[ name\t]') == [
+            'Attrib[Element[a][name]]']
+        assert parse_many('a [name]') == [
+            'CombinedSelector[Element[a] <followed> Attrib[Element[*][name]]]']
+        assert parse_many('a[rel="include"]') == [
+            "Attrib[Element[a][rel = String('include', 6)]]"]
+        assert parse_many('a[rel = include]') == [
+            "Attrib[Element[a][rel = Symbol('include', 8)]]"]
+        assert parse_many("a[hreflang |= 'en']") == [
+            "Attrib[Element[a][hreflang |= String('en', 14)]]"]
+        assert parse_many('div:nth-child(10)') == [
+            "Function[Element[div]:nth-child(Symbol('10', 14))]"]
+        assert parse_many(':nth-child(2n+2)') == [
+            "Function[Element[*]:nth-child(Symbol('2n+2', 11))]"]
+        assert parse_many('div:nth-of-type(10)') == [
+            "Function[Element[div]:nth-of-type(Symbol('10', 16))]"]
+        assert parse_many('div div:nth-of-type(10) .aclass') == [
             'CombinedSelector[CombinedSelector[Element[div] <followed> '
                 "Function[Element[div]:nth-of-type(Symbol('10', 20))]] "
-                '<followed> Class[Element[*].aclass]]')
-        assert parse_many('label:only') == (
-            'Pseudo[Element[label]:only]')
-        assert parse_many('a:lang(fr)') == (
-            "Function[Element[a]:lang(Symbol('fr', 7))]")
-        assert parse_many('div:contains("foo")') == (
-            "Function[Element[div]:contains(String('foo', 13))]")
-        assert parse_many('div#foobar') == (
-            'Hash[Element[div]#foobar]')
-        assert parse_many('div:not(div.foo)') == (
-            'Function[Element[div]:not(Class[Element[div].foo])]')
-        assert parse_many('td ~ th') == (
-            'CombinedSelector[Element[td] ~ Element[th]]')
+                '<followed> Class[Element[*].aclass]]']
+        assert parse_many('label:only') == [
+            'Pseudo[Element[label]:only]']
+        assert parse_many('a:lang(fr)') == [
+            "Function[Element[a]:lang(Symbol('fr', 7))]"]
+        assert parse_many('div:contains("foo")') == [
+            "Function[Element[div]:contains(String('foo', 13))]"]
+        assert parse_many('div#foobar') == [
+            'Hash[Element[div]#foobar]']
+        assert parse_many('div:not(div.foo)') == [
+            'Negation[Element[div]:not(Class[Element[div].foo])]']
+        assert parse_many('td ~ th') == [
+            'CombinedSelector[Element[td] ~ Element[th]]']
+
+    def test_pseudo_elements(self):
+        def parse_pseudo(css):
+            result = []
+            for selector in parse(css):
+                result.append((
+                    repr(selector._tree).replace("(u'", "('"),
+                    selector.pseudo_element))
+            return result
+
+        def parse_one(css):
+            result = parse_pseudo(css)
+            assert len(result) == 1
+            return result[0]
+
+        assert parse_one('foo') == ('Element[foo]', None)
+        assert parse_one('*') == ('Element[*]', None)
+        assert parse_one(':empty') == ('Pseudo[Element[*]:empty]', None)
+
+        # Special cases for CSS 2.1 pseudo-elements
+        assert parse_one(':before') == ('Element[*]', 'before')
+        assert parse_one(':after') == ('Element[*]', 'after')
+        assert parse_one(':first-line') == ('Element[*]', 'first-line')
+        assert parse_one(':first-letter') == ('Element[*]', 'first-letter')
+
+        assert parse_one('::before') == ('Element[*]', 'before')
+        assert parse_one('::after') == ('Element[*]', 'after')
+        assert parse_one('::first-line') == ('Element[*]', 'first-line')
+        assert parse_one('::first-letter') == ('Element[*]', 'first-letter')
+
+        assert parse_one('::selection') == ('Element[*]', 'selection')
+        assert parse_one('foo:after') == ('Element[foo]', 'after')
+        assert parse_one('foo::selection') == ('Element[foo]', 'selection')
+        assert parse_one('lorem#ipsum ~ a#b.c[href]:empty::selection') == (
+            'CombinedSelector[Hash[Element[lorem]#ipsum] ~ '
+                'Pseudo[Attrib[Class[Hash[Element[a]#b].c][href]]:empty]]',
+            'selection')
+
+        parse_pseudo('foo:before, bar, baz:after') == [
+            ('Element[foo]', 'before'),
+            ('Element[bar]', None),
+            ('Element[baz]', 'after')]
+
+    def test_specificity(self):
+        def specificity(css):
+            selectors = parse(css)
+            assert len(selectors) == 1
+            return selectors[0].specificity()
+
+        assert specificity('*') == (0, 0, 0)
+        assert specificity(' foo') == (0, 0, 1)
+        assert specificity(':empty ') == (0, 1, 0)
+        assert specificity(':before') == (0, 0, 1)
+        assert specificity('*:before') == (0, 0, 1)
+        assert specificity(':nth-child(2)') == (0, 1, 0)
+        assert specificity('.bar') == (0, 1, 0)
+        assert specificity('[baz]') == (0, 1, 0)
+        assert specificity('[baz="4"]') == (0, 1, 0)
+        assert specificity('[baz^="4"]') == (0, 1, 0)
+        assert specificity('#lipsum') == (1, 0, 0)
+
+        assert specificity('foo:empty') == (0, 1, 1)
+        assert specificity('foo:before') == (0, 0, 2)
+        assert specificity('foo::before') == (0, 0, 2)
+        assert specificity('foo:empty::before') == (0, 1, 2)
+
+        assert specificity('#lorem + foo#ipsum:first-child > bar:first-line'
+            ) == (2, 1, 3)
 
     def test_parse_errors(self):
         def get_error(css):
@@ -184,9 +264,24 @@ class TestCssselect(unittest.TestCase):
             "[Token('[', 0), Symbol('href', 1), Token(']', 5)]"
             " -> Symbol('a', 6)")
 
+        # Mis-placed pseudo-elements
+        assert get_error('a:before:empty') == (
+            "A pseudo-element must be at the end of a selector at "
+            "[Symbol('a', 0), Token(':', 1), Symbol('before', 2)] "
+            "-> Token(':', 8)")
+        assert get_error('li:before a') == (
+            "A pseudo-element must be at the end of a selector at "
+            "[Symbol('li', 0), Token(':', 2), Symbol('before', 3), "
+            "Token(' ', 9)] -> Symbol('a', 10)")
+        assert get_error(':not(:before)') == (
+            "Pseudo-elements are not allowed inside :not() at "
+            "[Token(':', 0), Symbol('not', 1), Token('(', 4), Token(':', 5),"
+            " Symbol('before', 6)] -> Token(')', 12)")
+
+
     def test_translation(self):
         def xpath(css):
-            return str(GenericTranslator().xpath(parse(css)))
+            return str(GenericTranslator().css_to_xpath(css, prefix=''))
 
         assert xpath('*') == "*"
         assert xpath('E') == "e"
@@ -334,6 +429,8 @@ class TestCssselect(unittest.TestCase):
         assert all_ids[:4] == ['html', 'nil', 'nil', 'outer-div']
         assert all_ids[-1:] == ['foobar-span']
         assert pcss('div') == ['outer-div', 'li-div', 'foobar-div']
+        assert pcss('div div') == ['li-div']
+        assert pcss('div, div div') == ['outer-div', 'li-div', 'foobar-div']
         assert pcss('a[name]') == ['name-anchor']
         assert pcss('a[rel]') == ['tag-anchor', 'nofollow-anchor']
         assert pcss('a[rel="tag"]') == ['tag-anchor']
