@@ -12,7 +12,7 @@
 """
 
 import re
-from cssselect.parser import parse, parse_series, Element, SelectorError
+from cssselect.parser import parse, parse_series, SelectorError
 
 
 try:
@@ -25,54 +25,34 @@ except NameError:
 
 
 class ExpressionError(SelectorError, RuntimeError):
-    """Inexistent or unsupported selector (eg. pseudo-class)."""
+    """Unknown or unsupported selector (eg. pseudo-class)."""
 
 
 #### XPath Helpers
 
 class XPathExpr(object):
 
-    def __init__(self, prefix=None, path=None, element='*', condition=None,
-                 star_prefix=False):
-        self.prefix = prefix
+    def __init__(self, path='', element='*', condition='', star_prefix=False):
         self.path = path
         self.element = element
         self.condition = condition
         self.star_prefix = star_prefix
 
     def __str__(self):
-        path = ''
-        if self.prefix is not None:
-            path += _unicode(self.prefix)
-        if self.path is not None:
-            path += _unicode(self.path)
-        path += _unicode(self.element)
+        path =  _unicode(self.path) + _unicode(self.element)
         if self.condition:
             path += '[%s]' % self.condition
         return path
 
     def __repr__(self):
-        return '%s[%s]' % (
-            self.__class__.__name__, self)
+        return '%s[%s]' % (self.__class__.__name__, self)
 
     def add_condition(self, condition):
         if self.condition:
             self.condition = '%s and (%s)' % (self.condition, condition)
         else:
             self.condition = condition
-
-    def add_path(self, part):
-        if self.path is None:
-            self.path = self.element
-        else:
-            self.path += self.element
-        self.element = part
-
-    def add_prefix(self, prefix):
-        if self.prefix:
-            self.prefix = prefix + self.prefix
-        else:
-            self.prefix = prefix
+        return self
 
     def add_name_test(self):
         if self.element == '*':
@@ -94,37 +74,17 @@ class XPathExpr(object):
         self.star_prefix = True
 
     def join(self, combiner, other):
-        prefix = _unicode(self)
-        prefix += combiner
-        path = (other.prefix or '') + (other.path or '')
+        path = _unicode(self) + combiner
         # We don't need a star prefix if we are joining to this other
         # prefix; so we'll get rid of it
-        if other.star_prefix and path == '*/':
-            path = ''
-        self.prefix = prefix
+        if not(other.star_prefix and other.path == '*/'):
+            path += other.path
         self.path = path
         self.element = other.element
         self.condition = other.condition
         return self
 
 
-class XPathExprOr(XPathExpr):
-    """
-    Represents |'d expressions.  Note that unfortunately it isn't
-    the union, it's the sum, so duplicate elements will appear.
-    """
-
-    def __init__(self, items, prefix=None):
-        for item in items:
-            assert item is not None
-        self.items = items
-        self.prefix = prefix
-
-    def __str__(self):
-        prefix = self.prefix or ''
-        return ' | '.join(["%s%s" % (prefix,i) for i in self.items])
-
-
 split_at_single_quotes = re.compile("('+)").split
 
 
@@ -152,49 +112,55 @@ class GenericTranslator(object):
         '!=': 'different',  # XXX Not in Level 3 but meh
     }
 
+    #: The attribute used for ID selectors depends on the document language:
+    #: http://www.w3.org/TR/selectors/#id-selectors
+    id_attribute = 'id'
+
     def css_to_xpath(self, css, prefix='descendant-or-self::'):
-        """Translate a CSS Selector to XPath.
+        """Translate a *group of selectors* to XPath.
 
-        :param css: An Unicode string or a parsed selector object.
-        :returns: An Unicode string.
+        Pseudo-elements are not supported here since XPath only knows
+        about "real" elements.
 
-        .. sourcecode:: pycon
-
-            >>> from cssselect import GenericTranslator
-            >>> expression = GenericTranslator().css_to_xpath('div.content')
-            >>> print(expression)
-            descendant-or-self::div[@class and contains(concat(' ', normalize-space(@class), ' '), ' content ')]
-
-        The resulting expression can be used with lxml's `XPath engine`_:
-
-        .. _XPath engine: http://lxml.de/xpathxslt.html#xpath
-
-        .. sourcecode:: pycon
-
-            >>> from lxml.etree import fromstring
-            >>> document = fromstring('''<div id="outer">
-            ...   <div id="inner" class="content body">
-            ...       text
-            ...   </div></div>''')
-            >>> [e.get('id') for e in document.xpath(expression)]
-            ['inner']
+        :param css:
+            A *group of selectors* as an Unicode string.
+        :raises:
+            :class:`SelectorSyntaxError` on invalid selectors,
+            :class:`ExpressionError` on unknown/unsupported selectors,
+            including pseudo-elements.
+        :returns:
+            The equivalent XPath 1.0 expression as an Unicode string.
 
         """
-        if isinstance(css, _basestring):
-            selector = parse(css)
-        else:
-            selector = css  # assume it is already parsed
-        xpath = self.xpath(selector)
-        xpath.add_prefix(prefix or '')
-        return _unicode(xpath)
+        selectors = parse(css)
+        for selector in selectors:
+            if selector.pseudo_element:
+                raise ExpressionError('Pseudo-elements are not supported.')
+
+        return ' | '.join(
+            self.selector_to_xpath(selector, prefix)
+            for selector in selectors)
+
+    def selector_to_xpath(self, selector, prefix='descendant-or-self::'):
+        """Translate a parsed selector to XPath.
+
+        The :attr:`~Selector.pseudo_element` attribute of the selector
+        is ignored. It is the caller's responsibility to reject selectors
+        with pseudo-elements, or to account for them somehow.
+
+        :param selector:
+            A parsed :class:`Selector` object.
+        :raises:
+            :class:`ExpressionError` on unknown/unsupported selectors.
+        :returns:
+            The equivalent XPath 1.0 expression as an Unicode string.
+
+        """
+        return (prefix or '') + _unicode(self.xpath(selector._tree))
 
     @staticmethod
     def xpath_literal(s):
-        if isinstance(s, Element):
-            # This is probably a symbol that looks like an expression...
-            s = s._format_element()
-        else:
-            s = _unicode(s)
+        s = _unicode(s)
         if "'" not in s:
             s = "'%s'" % s
         elif '"' not in s:
@@ -206,7 +172,7 @@ class GenericTranslator(object):
                 ])
         return s
 
-    def xpath(self, parsed_selector, prefix=None):
+    def xpath(self, parsed_selector):
         """Translate any parsed selector object."""
         type_name = type(parsed_selector).__name__
         method = getattr(self, 'xpath_%s' % type_name.lower(), None)
@@ -217,10 +183,6 @@ class GenericTranslator(object):
 
     # Dispatched by parsed object type
 
-    def xpath_or(self, selector_group):
-        """Translate a group of selectors (class Or)."""
-        return XPathExprOr([self.xpath(item) for item in selector_group.items])
-
     def xpath_combinedselector(self, combined):
         """Translate a combined selector."""
         combinator = self.combinator_mapping.get(combined.combinator)
@@ -231,6 +193,12 @@ class GenericTranslator(object):
         return method(self.xpath(combined.selector),
                       self.xpath(combined.subselector))
 
+    def xpath_negation(self, negation):
+        xpath = self.xpath(negation.selector)
+        sub_xpath = self.xpath(negation.subselector)
+        sub_xpath.add_name_test()
+        return xpath.add_condition('not(%s)' % sub_xpath.condition)
+
     def xpath_function(self, function):
         """Translate a functional pseudo-class."""
         method = 'xpath_%s_function' % function.name.replace('-', '_')
@@ -275,8 +243,8 @@ class GenericTranslator(object):
     def xpath_hash(self, id_selector):
         """Translate an ID selector."""
         xpath = self.xpath(id_selector.selector)
-        xpath.add_condition('@id = %s' % self.xpath_literal(id_selector.id))
-        return xpath
+        return xpath.add_condition('@%s = %s' % (
+            self.id_attribute, self.xpath_literal(id_selector.id)))
 
     def xpath_element(self, selector):
         """Translate a type or universal selector."""
@@ -302,8 +270,7 @@ class GenericTranslator(object):
         """right is a sibling immediately after left"""
         xpath = left.join('/following-sibling::', right)
         xpath.add_name_test()
-        xpath.add_condition('position() = 1')
-        return xpath
+        return xpath.add_condition('position() = 1')
 
     def xpath_indirect_adjacent_combinator(self, left, right):
         """right is a sibling after left, immediately or not"""
@@ -317,16 +284,14 @@ class GenericTranslator(object):
         a, b = parse_series(function.arguments)
         if not a and not b and not last:
             # a=0 means nothing is returned...
-            xpath.add_condition('false() and position() = 0')
-            return xpath
+            return xpath.add_condition('false() and position() = 0')
         if add_name_test:
             xpath.add_name_test()
         xpath.add_star_prefix()
         if a == 0:
             if last:
                 b = 'last() - %s' % b
-            xpath.add_condition('position() = %s' % b)
-            return xpath
+            return xpath.add_condition('position() = %s' % b)
         if last:
             # FIXME: I'm not sure if this is right
             a = -a
@@ -372,16 +337,9 @@ class GenericTranslator(object):
         return self.xpath_nth_child_function(xpath, function, last=True,
                                              add_name_test=False)
 
-    def xpath_not_function(self, xpath, function):
-        condition = self.xpath(function.arguments).condition
-        # FIXME: should I do something about element_path?
-        xpath.add_condition('not(%s)' % condition)
-        return xpath
-
     def xpath_contains_function(self, xpath, function):
-        xpath.add_condition('contains(string(.), %s)'
+        return xpath.add_condition('contains(string(.), %s)'
                             % self.xpath_literal(function.arguments))
-        return xpath
 
     def function_unsupported(self, xpath, pseudo):
         raise ExpressionError(
@@ -393,58 +351,49 @@ class GenericTranslator(object):
     # Pseudo: dispatch by pseudo-class name
 
     def xpath_root_pseudo(self, xpath):
-        xpath.add_condition("not(parent::*)")
-        return xpath
+        return xpath.add_condition("not(parent::*)")
 
     def xpath_first_child_pseudo(self, xpath):
         xpath.add_star_prefix()
         xpath.add_name_test()
-        xpath.add_condition('position() = 1')
-        return xpath
+        return xpath.add_condition('position() = 1')
 
     def xpath_last_child_pseudo(self, xpath):
         xpath.add_star_prefix()
         xpath.add_name_test()
-        xpath.add_condition('position() = last()')
-        return xpath
+        return xpath.add_condition('position() = last()')
 
     def xpath_first_of_type_pseudo(self, xpath):
         if xpath.element == '*':
             raise ExpressionError(
                 "*:first-of-type is not implemented")
         xpath.add_star_prefix()
-        xpath.add_condition('position() = 1')
-        return xpath
+        return xpath.add_condition('position() = 1')
 
     def xpath_last_of_type_pseudo(self, xpath):
         if xpath.element == '*':
             raise ExpressionError(
                 "*:last-of-type is not implemented")
         xpath.add_star_prefix()
-        xpath.add_condition('position() = last()')
-        return xpath
+        return xpath.add_condition('position() = last()')
 
     def xpath_only_child_pseudo(self, xpath):
         xpath.add_name_test()
         xpath.add_star_prefix()
-        xpath.add_condition('last() = 1')
-        return xpath
+        return xpath.add_condition('last() = 1')
 
     def xpath_only_of_type_pseudo(self, xpath):
         if xpath.element == '*':
             raise ExpressionError(
                 "*:only-of-type is not implemented")
-        xpath.add_condition('last() = 1')
-        return xpath
+        return xpath.add_condition('last() = 1')
 
     def xpath_empty_pseudo(self, xpath):
-        xpath.add_condition("not(*) and not(normalize-space())")
-        return xpath
+        return xpath.add_condition("not(*) and not(normalize-space())")
 
     def pseudo_never_matches(self, xpath):
         """Common implementation for pseudo-classes that never match."""
-        xpath.add_condition("0")
-        return xpath
+        return xpath.add_condition("0")
 
     xpath_link_pseudo = pseudo_never_matches
     xpath_visited_pseudo = pseudo_never_matches
@@ -492,22 +441,19 @@ class GenericTranslator(object):
         return xpath
 
     def xpath_attrib_prefixmatch(self, xpath, name, value):
-        xpath.add_condition('%s and starts-with(%s, %s)' % (
+        return xpath.add_condition('%s and starts-with(%s, %s)' % (
             name, name, self.xpath_literal(value)))
-        return xpath
 
     def xpath_attrib_suffixmatch(self, xpath, name, value):
         # Oddly there is a starts-with in XPath 1.0, but not ends-with
-        xpath.add_condition(
+        return xpath.add_condition(
             '%s and substring(%s, string-length(%s)-%s) = %s'
             % (name, name, name, len(value)-1, self.xpath_literal(value)))
-        return xpath
 
     def xpath_attrib_substringmatch(self, xpath, name, value):
         # Attribute selectors are case sensitive
-        xpath.add_condition('%s and contains(%s, %s)' % (
+        return xpath.add_condition('%s and contains(%s, %s)' % (
             name, name, self.xpath_literal(value)))
-        return xpath
 
 
 class HTMLTranslator(GenericTranslator):
@@ -516,21 +462,19 @@ class HTMLTranslator(GenericTranslator):
     """
     def xpath_checked_pseudo(self, xpath):
         # FIXME: is this really all the elements?
-        xpath.add_condition(
+        return xpath.add_condition(
             "(@selected and name(.) = 'option') or "
             "(@checked and name(.) = 'input')")
-        return xpath
 
     def xpath_link_pseudo(self, xpath):
-        xpath.add_condition("@href and name(.) = 'a'")
-        return xpath
+        return xpath.add_condition("@href and name(.) = 'a'")
 
     # Links are never visited, the implementation for :visited is the same
     # as in GenericTranslator
 
     def xpath_disabled_pseudo(self, xpath):
         # http://www.w3.org/TR/html5/section-index.html#attributes-1
-        xpath.add_condition('''
+        return xpath.add_condition('''
         (
             @disabled and
             (
@@ -557,11 +501,10 @@ class HTMLTranslator(GenericTranslator):
         ''')
         # FIXME: in the second half, add "and is not a descendant of that
         # fieldset element's first legend element child, if any."
-        return xpath
 
     def xpath_enabled_pseudo(self, xpath):
         # http://www.w3.org/TR/html5/section-index.html#attributes-1
-        xpath.add_condition('''
+        return xpath.add_condition('''
         (
             (
                 name(.) = 'command' or
@@ -583,4 +526,3 @@ class HTMLTranslator(GenericTranslator):
         ''')
         # FIXME: in the second half, add "and is not a descendant of that
         # fieldset element's first legend element child, if any."
-        return xpath
